<?php

declare(strict_types=1);

/*
 * Copyright (c) 2023 François Kooman <fkooman@tuxed.net>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

namespace fkooman\Radius\Tests;

use fkooman\Radius\Exception\AuthException;
use fkooman\Radius\Exception\RadiusException;
use PHPUnit\Framework\TestCase;

class RadiusClientTest extends TestCase
{
    public function testRequest(): void
    {
        $t = new TestSocket('02000040c5730d1a0db20e5d6ad5fb603102888e121a41757468656e7469636174696f6e2073756363656564656450127fecf9ed3c28ed5e203de65f98008402');
        $r = new TestRadiusClient($t, 'e38e27cdd0a5f45b801fa2510974fa1f', 'my-nas-id');
        $r->addServer('10.253.109.1', 's3cr3t', 1812);
        $accessResponse = $r->accessRequest('foo', 'bar');
        $this->assertSame(
            '01000048e38e27cdd0a5f45b801fa2510974fa1f0105666f6f0212a19fb3d0bb571a9059e868d3c2e90761200b6d792d6e61732d696450125f6221b6431e5188b39db84b915c96a8',
            bin2hex($t->getWriteBuffer())
        );
        $this->assertSame(2, $accessResponse->responseCodeValue());
        $this->assertSame('Access-Accept', $accessResponse->responseCode());
        $this->assertSame(
            [
                [18, 'Authentication succeeded'],
                [80, hex2bin('7fecf9ed3c28ed5e203de65f98008402')],
            ],
            $accessResponse->attributeList()
        );
        $this->assertSame('Authentication succeeded', $accessResponse->attributeByName('Reply-Message'));
        $this->assertSame('Authentication succeeded', $accessResponse->attributeByType(18));
        $this->assertSame(hex2bin('7fecf9ed3c28ed5e203de65f98008402'), $accessResponse->attributeByName('Message-Authenticator'));
    }

    public function testVeryLongPasswordRequest(): void
    {
        $t = new TestSocket('020000407a952b8823b81ebdf57b3812c4f19d1d121a41757468656e7469636174696f6e207375636365656465645012ca9ba652ebc9641d79fe860ea828b738');
        $r = new TestRadiusClient($t, 'd7e35415ff3b206ac917d32d823ae9a0', 'my-nas-id');
        $r->addServer('10.253.109.1', 's3cr3t', 1812);
        $accessResponse = $r->accessRequest('foo', str_repeat('a', 20));
        $this->assertSame(
            '01000058d7e35415ff3b206ac917d32d823ae9a00105666f6f022268688ff4b3fca0b188a1af2edbd88e2d3bfe53969c6af7ecd0af124ffb747edf200b6d792d6e61732d69645012f8799ab500a6aa81a18db4de69a4c52d',
            bin2hex($t->getWriteBuffer())
        );
        $this->assertSame(2, $accessResponse->responseCodeValue());
        $this->assertSame(
            [
                [18, 'Authentication succeeded'],
                [80, hex2bin('ca9ba652ebc9641d79fe860ea828b738')],
            ],
            $accessResponse->attributeList()
        );
    }

    public function testWrongPassword(): void
    {
        $t = new TestSocket('0300003d5e44868cf5ce3285c1defc6aa8082ed8121741757468656e7469636174696f6e206661696c65645012c934254f9b7dc45fb29c261582444a57');
        $r = new TestRadiusClient($t, '11bb163812653e088111c6a7ac6d8311', 'my-nas-id');
        $r->addServer('10.253.109.1', 's3cr3t', 1812);

        try {
            $r->accessRequest('foo', 'baz');
            $this->fail();
        } catch (AuthException $e) {
            $this->assertSame(
                '0100004811bb163812653e088111c6a7ac6d83110105666f6f02129db946a488988d569dad8266a01660f8200b6d792d6e61732d6964501299e5a8a0b124b503dc21208a2d479208',
                bin2hex($t->getWriteBuffer())
            );
            $accessResponse = $e->accessResponse();
            $this->assertSame(3, $accessResponse->responseCodeValue());
            $this->assertSame('Access-Reject', $accessResponse->responseCode());
            $this->assertSame(
                [
                    [18, 'Authentication failed'],
                    [80, hex2bin('c934254f9b7dc45fb29c261582444a57')],
                ],
                $accessResponse->attributeList()
            );
        }
    }

    public function testMangledResponseAuthenticator(): void
    {
        $this->expectException(RadiusException::class);
        $this->expectExceptionMessage('Response Authenticator has an unexpected value');
        // we change the last two bytes from 02 -> 01 in the response
        // authenticator, just to see it fail
        $t = new TestSocket('02000040c5730d1a0db20e5d6ad5fb603102888e121a41757468656e7469636174696f6e2073756363656564656450127fecf9ed3c28ed5e203de65f98008401');
        $r = new TestRadiusClient($t, 'e38e27cdd0a5f45b801fa2510974fa1f', 'my-nas-id');
        $r->addServer('10.253.109.1', 's3cr3t', 1812);
        $r->accessRequest('foo', 'bar');
    }
}
